%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%  zip3([],[],[]) --> []
%%  zip3([X1|R1],[X2|R2],[X3|R3]) --> [tupla(X1,X2,X3)|zip3(R1,R2,R3)]
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
zip3(X1,X2,X3,H):-hnf(X1,HX1),zip3_1(HX1,X2,X3,H).

zip3_1([],X2,X3,H):-hnf(X2,HX2), zip3_1_nil_2(HX2,X3,H).
zip3_1_nil_2([],X3,H):-hnf(X3,HX3), zip3_1_nil_2_nil_3(HX3,H).
zip3_1_nil_2_nil_3([],[]).

zip3_1([A|B],X2,X3,H):-hnf(X2,HX2),zip3_1_cons_2(HX2,A,B,X3,H).
zip3_1_cons_2([C|D],A,B,X3,H):-hnf(X3,HX3),zip3_1_cons_2_cons_3(HX3,C,D,A,B,H).
zip3_1_cons_2_cons_3([E|F],C,D,A,B,[tuple(A,C,E)|zip3(B,D,F)]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block IMPROVED zip3.
izip3(A,B,C,G):-hnf(A,D),hnf(B,E),hnf(C,F),izip3_1(D,E,F,G).

izip3_1([],[],[],[]).
izip3_1([A|B],[C|D],[E|F],[tuple(A,C,E)|izip3(B,D,F)]).





%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(zip3(A,B,C),D):-!,zip3(A,B,C,D).
hnf(izip3(A,B,C),D):-!,izip3(A,B,C,D).
hnf(A,A).

%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf(tuple(A,B,C),tuple(D,E,F)):-!,nf(A,D),nf(B,E),nf(C,F).
nf_hnf([],[]):-!.
nf_hnf([A|B],[C|D]):-!,nf(A,C),nf(B,D).
nf_hnf(a,a):-!.
nf_hnf(b,b):-!.
nf_hnf(c,c):-!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
test(ZIP, IZIP) :- nf(zip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b]), ZIP),
      nf(izip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b]), IZIP).



%%%/**************** SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- nf(Computation,_), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start, g(V,_), stat_end(T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: zip3  ..'), stat_start,
	g(V,zip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b])), stat_end(T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: izip3 ..'), stat_start,
	g(V,izip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b])), stat_end(T).

allgoals:- V = 1000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.
%%%*******************************************************/

/**************** SWI-ROLOG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% T1 is the start time
stat_start(T1) :- T1 is cputime.

%% stat_end(Input,Output)
%% T1 is the starting time and T is the temporal cost of the whole computation,
stat_end(T1,T) :-
	T2 is cputime, T is T2-T1, write('T= '), write(T),
        nl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- nf(Computation,_), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start(T1), g(V,_), stat_end(T1,T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: zip3  ..'), stat_start(T1),
	g(V,zip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b])), stat_end(T1,T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: izip3 ..'), stat_start(T1),
	g(V,izip3([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b],[a,b,c,a,b,c,b,b,b])), stat_end(T1,T).

allgoals:- V = 100000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.

**********************************************************/





