%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%  zip([],[]) --> []
%%  zip([X1|R1],[X2|R2]) --> [tupla(X1,X2)|zip(R1,R2)]
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% :- block zip(?,?,?,-,?).
zip(A,B,E):-hnf(A,F),zip_1(F,B,E).

%% :- block zip_1(-,?,?,?,?),zip_1(?,?,?,-,?).
zip_1([],A,H):-hnf(A,B), zip_1_nil_1(B,H).
zip_1_nil_1([],[]).

zip_1([A|B],C,F):-hnf(C,G),'zip_1_._3'(G,A,B,F).

%% :- block 'zip_1_._3'(-,?,?,?,?,?),'zip_1_._3'(?,?,?,?,-,?).
'zip_1_._3'([A|B],C,D,[tuple(C,A)|zip(D,B)]).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block IMPROVED zip.
izip(A,B,E):-hnf(A,C),hnf(B,D),izip_1(C,D,E).

izip_1([],[],[]).
izip_1([A|B],[C|D],[tuple(A,C)|izip(B,D)]).




%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(zip(A,B),C):-!,zip(A,B,C).
hnf(izip(A,B),C):-!,izip(A,B,C).
hnf(A,A).

%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf(tuple(A,B),tuple(C,D)):-!,nf(A,C),nf(B,D).
nf_hnf([],[]):-!.
nf_hnf([A|B],[C|D]):-!,nf(A,C),nf(B,D).
nf_hnf(a,a):-!.
nf_hnf(b,b):-!.
nf_hnf(c,c):-!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
test(ZIP, IZIP) :- nf(zip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b]), ZIP),
      nf(izip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b]), IZIP).



%%%/**************** SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- nf(Computation,_), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start, g(V,_), stat_end(T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: zip  ..'), stat_start,
	g(V,zip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b])), stat_end(T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: izip ..'), stat_start,
	g(V,izip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b])), stat_end(T).

allgoals:- V = 10000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.
%%%*******************************************************/

/**************** SWI-ROLOG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% T1 is the start time
stat_start(T1) :- T1 is cputime.

%% stat_end(Input,Output)
%% T1 is the starting time and T is the temporal cost of the whole computation,
stat_end(T1,T) :-
	T2 is cputime, T is T2-T1, write('T= '), write(T),
        nl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- nf(Computation,_), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start(T1), g(V,_), stat_end(T1,T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: zip  ..'), stat_start(T1),
	g(V,zip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b])), stat_end(T1,T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: izip ..'), stat_start(T1),
	g(V,izip([a,b,c,c,b,a,a,a,a],[a,b,c,a,b,c,b,b,b])), stat_end(T1,T).

allgoals:- V = 100000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.

**********************************************************/


